/****************************************************
Generic parameter reconstructor main-including file.
This is the program for the paper:
"Rigorous parameter-reconstruction for differential
equations with noisy data."

Authors: Tomas Johnson    Uppsala University 
         Warwick Tucker   University of Bergen


***************************************************/


#include <ostream> 
#include <fstream>
#include "paramReconstruct.h"
#include "vnode.h"
#include "list.h"
#include "utilities.h"
using namespace std;
using namespace vnodelp;

/*
Enter your vector field into this template function, in the form
yp=f_i(y,p), where yp denotes \dot{y}.
The parameters should be supplied as an iVector.
For more information about the syntax, see the VNODE-LP manual.
 */


template<typename var_type> 
void PODE(int n,var_type *yp,const var_type *y,var_type t, void *param) {
  iVector *p=(iVector *) param;
  yp[0]= ; //First component of your vector field.
  yp[1]=  ;//Second component etc.
}
 

/*This function generates a test case.
The parameters r-eps and a-eps, denote the 
relative and absolute noise levels, respectively.
Insert the original parameter value from which the
trajectories are to be generated into the vector 'p'.
The initial value should be entered into the vector y[0][i].
Several trajectories can be supplied, simply copy the code which 
generates the trajectory and re-put it after the 'yList+=' command.
The trajectories are put as matrices in yList.
*/
void generateTest(int n, int k, int l, List<iMatrix> &yList, iVector &time) {
  interval t= START_TIME,tend=END_TIME;  //Start and end times
  interval h=(tend-t)/(k-1);  //Sample distance
  double reps= ;  //The relative noise level 
  double aeps= ;  //The absolute noise level

  cout << "Relative Epsilon: " << reps << ", " << endl
       << "Absolute Epsilon: " << aeps << "." << endl;

  interval error(1-reps,1+reps);
  iVector abserr(n);
  setV(abserr,interval(-aeps,aeps));

  //Insert the parameters that should be used to generate the data
  iVector p(l);
  p[0]=interval();
  //...
  p[l-1]=interval();
  
  AD *ad= new FADBAD_AD(n, PODE, PODE , &p);
  VNODE *Solver= new VNODE(ad);
  ad->eval(&p);
  iMatrix y(k,n);

  //------------------------------------------------
  //Insert the initial value
  y[0][0]=interval(INIT_VALUE_1);
  //...
  y[0][n-1]=interval(INIT_VALUE_n);
  
  Solver->setHmin(0);
  Solver->setFirstEntry();
  time[0]=t;

  for(int j=1;j<k;j++) {
    time[j]=time[j-1]+h;
    y[j]=y[j-1];
    t=time[j-1];
 
    Solver->integrate(t,y[j],time[j]); //Integrate !
    addViVi(y[j-1],abserr);  //Add the absolute error
    scaleV(y[j-1], error); //Add the relative error
    //If information about the phase space (e.g. positivity)
    //is known, insert it here.
    //   for(int u=0;u<n;u++)
    //      intersect(y[j-1][u],y[j-1][u],interval(0,Machine::PosInfinity));
  }
  addViVi(y[k-1],abserr);
  scaleV(y[k-1], error);
  //Positivity
  //   for(int u=0;u<n;u++)
  //      intersect(y[k-1][u],y[k-1][u],interval(0,Machine::PosInfinity));
 
  yList+=y; //Add one trajectory
  //------------------------------------------------
  //If generating with several trajectories. Insert between --- and --- here.

  cerr<<"Data generated with "<<Length(yList)<<" initial values."<<endl;
}

int main(){
  int n= ;   //The number of phase (state) space dimensions
  int k= ;   //The number of sample points
  int l= ;   //The number of parameters
 
  double tol=Power(2,-TOL); //Splitting tolerance
  double tol2=Power(2,-TOL2); //Pre-split tolerance

  iVector param(l);
  pVector weights(l);

  //param should contain the search domain for the unknown parameters.
  param[0]=interval(INF, SUP); //First parameter domain
  //...
  param[l-1]=  ; //The final parameter domain

  for(int j=0 ;j<l;j++)  //Normalise the metric
    weights[j]=1/width(param[j]);
  
  iVector time(k);
  List<iMatrix> yList;

  generateTest(n,k,l,yList,time);
 
  Tfunction ODE;
  TFfunction ODE2;
  ODE=PODE;
  ODE2=PODE;
  clock(START_TIMING);
  paramReconstruct(ODE, ODE2, tol, tol2, n, k, l, weights, param, time, yList);

  clock(STOP_TIMING);
  return 0;
}
