/****************************************************
Generic parameter reconstructor.
This file contains the reconstructing algorithms.

This is the program for the paper:
"Rigorous parameter-reconstruction for differential
equations with noisy data."

Authors: Tomas Johnson    Uppsala University 
         Warwick Tucker   University of Bergen


****************************************************/
#include "paramReconstruct.h"


//A struct containing a search domain for the parameters
//and a list of known trajcetories.

class paramAndy {
 public:
  List<iMatrix> yList;
  iVector param;
};

//The procedure reduce is a State space estimator.
//It integrates from one sample point to the next
//and intersects the result with the next one.
//Returns false if an inconsistency is found.
bool reduce(const int &n, iVector &yfrom, iVector &yto, interval &tfrom, interval &tto, VNODE *Solver) {
  Solver->setFirstEntry();
  iVector y=yfrom;
  interval t=tfrom;
  Solver->setHmin(0);
  if(!Solver->successful())
    return 1;
  Solver->setFirstEntry();
  Solver->integrate(t,y,tto);
  if(!Solver->successful())
    return 1;
  if(!intersect(yto,yto,y))
    return 0;
  
  return 1;
}

//Midway reduction. Flows a sample point to the midpoint (time-wise)
//of itself and the next sample point, and vice versa, and intersecs
//the resulting boxes.
//Returns false if an inconsistency is found.

bool reduce2(const int &n, iVector &yfrom, iVector &yto, interval &tfrom, interval &tto, VNODE *Solver) {
  Solver->setFirstEntry();
  iVector y1=yfrom;
  interval t1=tfrom;
  iVector y2=yto;
  interval t2=tto;
  interval tmid=(tfrom+tto)/2;

  Solver->setHmin(0);
  if(!Solver->successful())
	return 1;

  Solver->setFirstEntry();
  Solver->integrate(t1,y1,tmid);
  if(!Solver->successful())
    return 1;
  Solver->setHmin(0);
  if(!Solver->successful())
	return 1;
  Solver->setFirstEntry();
  Solver->integrate(t2,y2,tmid);
  if(!Solver->successful())
    return 1;

  if(!intersect(y1,y1,y2))
    return 0;
  return 1;
}

//Check consistency, returns 0(false) if one of the tests fails, i.e. if the
//given paramter domain is inconsistent with the data.
bool consistent(const int &n, const int &k, iVector *time, iMatrix &y, VNODE *Solver) {
  for(int i=1; i<k; i++) {
    if(!reduce(n, y[i-1], y[i], (*time)[i-1], (*time)[i], Solver))
      return 0;   
  }
  
  for(int i=k-1; i>0; i--) {
    if(!reduce(n, y[i], y[i-1], (*time)[i], (*time)[i-1], Solver))
        return 0;
    } 

  for(int i=1; i<k; i++) {
    if(!reduce2(n, y[i-1], y[i], (*time)[i-1], (*time)[i], Solver))
      return 0;
  }

  return 1;
}

//Determines which parameter in the domain the should be split.
void splitIndex(const int &l,const iVector &param, const pVector &weights, int &i) {
  i=0;
  double test=weights[0]*width(param[0]);
  for(int j=1;j<l;j++) {
    if( weights[j]*width(param[j])>test){
    i=j;
    test=weights[j]*width(param[j]);
    }
  }
}

//Checks if a paramter box is smaller than the given tolerance limit.
double checkTol(const int &l, const iVector &param, const pVector &weights) {
  double tol=weights[0]*width(param[0]);
  for(int i=1;i<l;i++)
    tol*=weights[i]*width(param[i]);
  return tol;
}

//If necessary pre-split the parameter domain, until the size
//requirement given by the tolerance tol2 , is satisfied.
void preSplitParam(const double &tol2, const int &l, const pVector &weights, const iVector &param, List<iVector> &paramList) { 	    
  if (checkTol(l, param, weights)>tol2) {
    int rIndex;
    splitIndex(l, param, weights, rIndex );
    iVector paramOut=param;
    interval chInt=param[rIndex];
    double infIndex=inf(chInt);
    double midIndex=midpoint(chInt);
    double supIndex=sup(chInt);
    
    paramOut[rIndex]=interval(infIndex,midIndex);
    preSplitParam(tol2, l, weights, paramOut, paramList);
    paramOut[rIndex]=interval(midIndex,supIndex);
    preSplitParam(tol2, l, weights, paramOut, paramList);
  }
  else
    paramList+=param;
}

//Calls the pre-splitter.
void generateParam(const double tol2, int n, int l, const pVector weights, const iVector &param, List<iVector> &paramList) {
  double mass=weights[0]*width(param[0]);
  for(int j=1; j<l; j++)
    mass*=weights[j]*width(param[j]);
  cout<<"initial mass = "<<mass<<endl;
  
  preSplitParam(tol2, l, weights, param, paramList);
}


//Investigate if a parameter should be considered to be consistent with the data.
bool acceptParam(const int &n,const int &k, iVector *time, List<iMatrix> *yList, VNODE *Solver) {
  
  pVector ymid(n);
  iMatrix yLoc=First(*yList);
  int checklen=Length(*yList);
  iVector yCheck=yLoc[0];

  for(int j=1;j<=checklen;j++) {
    midpoint(ymid,yCheck);  //Extract the midpoint of the state space
    iVector y(n);
    setV(y,0);
    addViVp(y,ymid);
    Solver->setHmin(0);
    if(!Solver->successful())
       return false;
    Solver->setFirstEntry();
    
    for(int i=0;i<k-1;i++) {
      interval t=(*time)[i];
      interval tto=(*time)[i+1];
      Solver->integrate(t,y,tto);
      if(!Solver->successful())
	return false;
      if(!subseteq(y,yLoc[i+1]))  //Check inclusion in the next sample interval
	return false;
    }
    
    if (j<checklen) {
      yLoc=Next(*yList);
      yCheck=yLoc[0];
    }
  }
 
  return true;
}


//Take a parameter/state part and check consistency and inconsistency. If none of them
//is satisfied, split the parameter search domain.
void splitAndCheck(const int &n,const int &k, const double &tol, const int &l, const pVector &weights, iVector *time, 
		   iVector checkParam, List<iVector> *YesList, List<iVector> *MaybeList, 
		   List<iVector> *NoList, List<iMatrix> *yOrgList, AD *ad, VNODE *Solver) {
  List<paramAndy> paramList;
  paramAndy checkThis;
  checkThis.param=checkParam;
  checkThis.yList=*yOrgList;
  paramList+=checkThis;
  while(!IsEmpty(paramList)) {
    checkThis=Pop(paramList);
    ad->eval(&checkThis.param);
   
    //Check consistency, (of one trajectory) !
    if (acceptParam(n,k, time, &checkThis.yList, Solver)) 
      (*YesList)+=checkThis.param;  //Consistent parameters.
    else
      { 
      iMatrix yCheck;
      int checklen=Length(*yOrgList);
      int i=1;
      yCheck=Pop(checkThis.yList);
            
      // Check inconsistency, (of all trajectories) !
      while( i<=checklen && consistent(n,k, time, yCheck, Solver)) {
	checkThis.yList*=yCheck;
	if (i<checklen)
	  yCheck=Pop(checkThis.yList);
	i++;
      }
     
      //Split !
      if(i>checklen) { 
	if (checkTol(l, checkThis.param, weights)>tol) {
	  int rIndex;
	  splitIndex(l, checkThis.param, weights, rIndex );
	  interval chInt=checkThis.param[rIndex];
	  double infIndex=inf(chInt);
	  double midIndex=midpoint(chInt);
	  double supIndex=sup(chInt);
	  
	  checkThis.param[rIndex]=interval(infIndex,midIndex);
	  paramList*=checkThis;
	  checkThis.param[rIndex]=interval(midIndex,supIndex);
	  paramList*=checkThis;
	}
	else {
	  (*MaybeList)+=checkThis.param;  //Parameters that are undecided
	                                  //After the tolerance, tol
 	}                                 //is violated
      }
      else 
	(*NoList)+=checkThis.param;       //Inconsistent paramters.

    }
  } 
}

void paramReconstruct( Tfunction ODE, TFfunction ODE2, const double &tol, const double &tol2, const int &n, const int &k, const int &l,
		      const pVector &weights, iVector checkParam, iVector &time, List<iMatrix> yList) {
  cout << "Tolerance: " << tol << "." << endl;
  cout << "Using " << k << " samples." << endl;

  List<iVector> YesList, NoList, MaybeList;
  List<iVector> paramList;

  generateParam(tol2, n,l, weights, checkParam, paramList); //Pre-split
  checkParam=First(paramList);
  AD *ad= new FADBAD_AD(n, ODE, ODE2, &checkParam);
  VNODE *Solver= new VNODE(ad);

  cout << "Parameter region split into " << Length(paramList) << " boxes"<<endl;
  while(!IsEmpty(paramList)) {
    cout<< "|paramList| = " << Length(paramList) <<endl;
    checkParam=Pop(paramList);
    //Call the reconstructor
    splitAndCheck(n,k, tol, l, weights, &time, checkParam, &YesList, &MaybeList, &NoList, &yList, ad, Solver);
  }
  
  //The rest of the function handles the output of the program.
  int len = Length(NoList);
  cout << "|NoList| = " << len << endl;
  ofstream OutFile("No.txt", ios_base::out);
  for(int i=1; i<=len; i++) {
    iVector param=Pop(NoList);
    for(int j=0;j<l;j++)
      OutFile << inf(param[j]) << " " << sup(param[j]) << " ";
	
    OutFile<<endl;
  }	    
  OutFile.close();
  len = Length(MaybeList);
  cout << "|MaybeList| = " << len << endl;
  OutFile.open("Maybe.txt");
  pVector centerOfMass(l);
  setV(centerOfMass,0);
  iVector enclosure(l), Yenclosure(l);
  double mass, totmass=0, Ytotmass=0;
  iVector param(l);

  if (!IsEmpty(MaybeList)) {
    param=First(MaybeList);
    enclosure=param;
    
    for(int i=1; i<=len; i++) {
      param=Pop(MaybeList);
      
      for(int j=0;j<l;j++){
	enclosure[j]=Hull(enclosure[j],param[j]);
	OutFile << inf(param[j]) << " " << sup(param[j]) << " ";
      }
      OutFile<<endl;
      mass=width(param[0])*weights[0];
      for(int j=1;j<l;j++)
	mass*=width(param[j])*weights[j];
      totmass+=mass;
      for(int j=0;j<l;j++)
	centerOfMass[j]+=midpoint(param[j])*mass;
    }
  }
  OutFile.close();
  if (!IsEmpty(YesList)) {
    OutFile.open("Yes.txt");
    len = Length(YesList);
    cout << "|YesList| = " << len << endl;
    param=First(YesList);
    Yenclosure=param;

    for(int i=1; i<=len; i++) {
      param=Pop(YesList);
      
      for(int j=0;j<l;j++){
	enclosure[j]=Hull(enclosure[j],param[j]);
	Yenclosure[j]=Hull(Yenclosure[j],param[j]);
	OutFile << inf(param[j]) << " " << sup(param[j]) << " ";
      }
      OutFile<<endl;
      mass=width(param[0])*weights[0];
      for(int j=1;j<l;j++)
	mass*=width(param[j])*weights[j];
      totmass+=mass;
      Ytotmass+=mass;
      for(int j=0;j<l;j++)
	centerOfMass[j]+=midpoint(param[j])*mass;
    }
    OutFile.close();
  }
  if(totmass==0)
    cout<<"No parameters are consistent with the data! "<<endl;
  else {
    cout<<"weighted average: "<<endl;
    scaleV(centerOfMass, 1/totmass);
    printVector(centerOfMass);
    cout<<"Total remaining mass = " << totmass<<endl;
    cout<<"Consistent remaining mass = " << Ytotmass<<endl;
    cout<<"Enclosures : "<<endl;
    printVector(enclosure);
    cout<<"Consistent enclosures : "<<endl;
    printVector(Yenclosure);
  }
}

